(*
 * lisp.ml
 *
 * ocamlscheme
 *
 * Copyright (C) 2001-2004 Mark Probst
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *)

open Type
open Symtab
open Format
open List
exception Builtin_Invalid_Args of string
exception Print_Error
exception Not_Symbol
exception Not_Value
exception Compile_Error
exception Apply_Error
exception Expand_Error

(*******************************************************
	PRINTING FOR DEBUGGING PURPOSES
*******************************************************)
let rec prindebug x =
  match x with
    Nil -> print_string "nil ()"
  | Int i -> printf "int %i" i
  | Float f -> printf "float %f" f
  | String s -> printf "string %s" s
  | Symbol { name = n } -> printf "symbol %s" n
  | Cons(car, cdr) -> print_string "cons ("; prindebug car; print_string " . "; prindebug cdr; print_string ")"
  | _ -> raise Print_Error;;

let rec prinlist x =
  match x with
    Cons(car, Cons(cadr, cddr)) -> prin1 car; print_string " "; prinlist(Cons(cadr, cddr))
  | Cons(car, Nil) -> prin1 car
  | Cons(car, cdr) -> prin1 car; print_string " . "; prin1 cdr
  | _ -> raise Print_Error
and prin1 x =
  match x with
    Nil -> print_string "()"
  | Int i -> print_int i
  | Float f -> print_float f
  | String s -> print_string s
  | Symbol { name = n } -> print_string n
  | Cons(car, cdr) -> print_string "("; prinlist(Cons(car, cdr)); print_string ")"
  | Builtin _ -> print_string "#<builtin>"
  | Func _ -> print_string "#<func>"
  | Closure _ -> print_string "#<closure>"
  | Compiled_closure _ -> print_string "#<compiled_closure>";;

  
(*****************************************************
	SYMBOLS , ENVIRONMENT , COMPILATION , EVALUATION
******************************************************)  
  
let rec uncons_symbols x =
  match x with
    Nil -> []
  | Cons(Symbol s, cdr) -> s :: uncons_symbols cdr
  | _ -> raise Not_Symbol;;
  
let nil_symbol = intern "#nil"
and t_symbol = intern "#t";;

let symbol_value_value s =
  match s.value with
    Value v -> v
  | _ -> raise Not_Value;;

let rec position f l =
  match l with
    [] -> raise Not_found
  | x :: r -> if f x then 0 else 1 + position f r;;

  
let rec env_lookup name e =
  match e with
    [] -> raise Not_found
  | a :: r ->
      try
	(0, position (function x -> x = name) a)
      with Not_found ->
	match env_lookup name r with
	  (depth, index) -> (depth + 1, index);;

let rec compile x e =
  match x with
    Nil -> Quote x
  | Int _ -> Quote x
  | Float _ -> Quote x
  | String _ -> Quote x
  | Builtin _ -> Quote x
  | Func _ -> Quote x
  | Closure _ -> Quote x
  | Compiled_closure _ -> Quote x
  | Symbol s ->
      (
       try
	 match env_lookup s e with
	   (depth, index) -> Var (depth, index)
       with Not_found ->
	 Global s
      )
  | Cons(Symbol { name = "quote" }, Cons(x, Nil)) -> Quote(x)
  | Cons(Symbol { name = "if" }, Cons(cond, Cons(cons, Cons(alt, Nil)))) -> If (compile cond e, compile cons e, compile alt e)
  | Cons(Symbol { name = "lambda" }, Cons(args, Cons(body, Nil))) -> Build_closure (compile body ((uncons_symbols args) :: e))
  | Cons(Symbol { name = "define" }, Cons(Symbol { name = name }, Cons(value, Nil))) -> Global_set (global_lookup name, compile value e)
  | Cons(Symbol { name = "define-macro" }, Cons(Symbol { name = name }, Cons(value, Nil))) -> Global_set_macro (global_lookup name, compile value e)
  | Cons(Symbol { name = "apply" }, Cons(func, args)) -> Application (compile func e, Array.of_list (compile_args args e))
  | Cons(func, args) -> Application (compile func e, Array.of_list (compile_args args e))
and compile_args a e =
  match a with
    Nil -> []
  | Cons(arg, rest) -> compile arg e :: compile_args rest e
  | _ -> raise Compile_Error;;

let rec bcapply f a =
  match f with
    Builtin b -> b a
  | Closure(c, e) -> bceval c (a :: e)
  | _ -> raise Apply_Error
and bceval x e =
  match x with
    Quote q -> q
  | Global s -> symbol_value_value s
  | If (cond, cons, alt) -> if bceval cond e == Nil then bceval alt e else bceval cons e
  | Application (f, args) -> bcapply (bceval f e) (Array.map (function a -> bceval a e) args)
  | Build_closure c -> Closure (c, e)
  | Global_set (s, c) -> s.value <- Value (bceval c e) ; Nil
  | Global_set_macro (s, c) -> s.value <- Macro (bceval c e) ; Nil
  | Var (depth, index) -> (nth e depth).(index);;

let rec macroexpand_list x =
  match x with
    Nil -> []
  | Cons (car, cdr) -> macroexpand car :: macroexpand_list cdr
  | _ -> raise Expand_Error
and macroexpand x =
  match x with
    Cons (Symbol { value = Macro m }, args) -> (bcapply m (Array.of_list (macroexpand_list args)))
  | Cons (car, cdr) -> Cons (macroexpand car, macroexpand cdr)
  | _ -> x;;


  
(*****************************************************
	BUILTINS: built-in functions
******************************************************)  

(*************** Arithmetic and logic *********************)
  
(* function '+' *)
let builtin_plus args =
  Array.fold_left (function x -> function y -> 
  						match (x, y) with 
  						(Int(i), Int(j)) -> Int(i + j) 
  					|	(Float(i), Int(j)) -> Float(i +. (float_of_int j))
  					|	(Int(i), Float(j)) -> Float(j +. (float_of_int i))
  					|	(Float(i), Float(j)) -> Float(i +. j)
  					| _ -> raise (Builtin_Invalid_Args "Invalid arguments to '+'. All arguments must be numbers.\n")) (Int(0)) args;;

(* function '-' *)
let rec builtin_minus args =
  match args.(0) with
    Int(i) ->
      (
       match builtin_plus(args) with
	 	Int(j) -> Int(i + i - j)
	   |Float(j) -> Float((float_of_int i) +. (float_of_int i) -. j)
       | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '-'. All arguments must be numbers.\n")
      )
  | Float(i) ->   
      (
       match builtin_plus(args) with
	 	Int(j) -> Float(i +. i -. (float_of_int j))
	   |Float(j) -> Float(i +. i -. j)
       | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '-'. All arguments must be numbers.\n")
      )
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '-'. All arguments must be numbers.\n");;
  
(* function '*' *)
let builtin_times args =
  Array.fold_left (function x -> function y -> 
  						match (x, y) with 
  						(Int(i), Int(j)) -> Int(i * j) 
  					|	(Float(i), Int(j)) -> Float(i *. (float_of_int j))
  					|	(Int(i), Float(j)) -> Float(j *. (float_of_int i))
  					|	(Float(i), Float(j)) -> Float(i *. j)
  					| _ -> raise (Builtin_Invalid_Args "Invalid arguments to '*'. All arguments must be numbers.\n")) (Int(1)) args;;
  					
(* function '/' *)
let builtin_divide args =
  match args with
    [|Int(i); Int(j)|] -> Int(i / j)
  | [|Int(i); Float(j)|] -> Float( (float_of_int i) /. j)
  | [|Float(i); Int(j)|] -> Float( i /. (float_of_int j))
  | [|Float(i); Float(j)|] -> Float( i /. j)
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '/'. Both arguments must be numbers.\n");; 					

(* function 'quotient' *)
let builtin_quotient args =
  match args with
    [|Int(i); Int(j)|] -> Int(i / j)
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to 'quotient'. Both arguments must be integers.\n");; 					

(* function 'modulo' *)
let builtin_modulo args =
  match args with
    [|Int(i); Int(j)|] -> Int(i mod j)
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to 'modulo'. Both arguments must be integers.\n");; 					
    
  
(* function '=' *)  					  
let builtin_numbers_equal args =
  match args with
    [|Int(i); Int(j)|] -> if i = j then Symbol t_symbol else Nil
  | [|Int(i); Float(j)|] -> if (float_of_int i) = j then Symbol t_symbol else Nil
  | [|Float(i); Int(j)|] -> if (float_of_int j) = i then Symbol t_symbol else Nil
  | [|Float(i); Float(j)|] -> if i = j then Symbol t_symbol else Nil
  | [|String(i); String(j)|] -> if i = j then Symbol t_symbol else Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '='. Both arguments must be the same type.\n");;
  
(* function 'not' *)  
let builtin_numbers_not args =
  match args with
   [|Nil|] -> Symbol t_symbol
  | [|t_symbol|] -> Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid argument to 'not'. Argument must be boolean.\n");;
  
(* function '<' *) 
let builtin_less args =
  match args with
    [|Int(i); Int(j)|] -> if i < j then Symbol t_symbol else Nil
  | [|Int(i); Float(j)|] -> if (float_of_int i) < j then Symbol t_symbol else Nil
  | [|Float(i); Int(j)|] -> if i < (float_of_int j) then Symbol t_symbol else Nil
  | [|Float(i); Float(j)|] -> if i < j then Symbol t_symbol else Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '<'. Both arguments must be numbers.\n");;

(* function '<=' *) 
let builtin_lesseq args =
  match args with
    [|Int(i); Int(j)|] -> if i <= j then Symbol t_symbol else Nil
  | [|Int(i); Float(j)|] -> if (float_of_int i) <= j then Symbol t_symbol else Nil
  | [|Float(i); Int(j)|] -> if i <= (float_of_int j) then Symbol t_symbol else Nil
  | [|Float(i); Float(j)|] -> if i <= j then Symbol t_symbol else Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '<='. Both arguments must be numbers.\n");;
  
 
(* function '>' *) 
let builtin_more args =
  match args with
    [|Int(i); Int(j)|] -> if i > j then Symbol t_symbol else Nil
  | [|Int(i); Float(j)|] -> if (float_of_int i) > j then Symbol t_symbol else Nil
  | [|Float(i); Int(j)|] -> if i > (float_of_int j) then Symbol t_symbol else Nil
  | [|Float(i); Float(j)|] -> if i > j then Symbol t_symbol else Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '>'. Both arguments must be numbers.\n");;

(* function '>=' *) 
let builtin_moreeq args =
  match args with
    [|Int(i); Int(j)|] -> if i >= j then Symbol t_symbol else Nil
  | [|Int(i); Float(j)|] -> if (float_of_int i) >= j then Symbol t_symbol else Nil
  | [|Float(i); Int(j)|] -> if i >= (float_of_int j) then Symbol t_symbol else Nil
  | [|Float(i); Float(j)|] -> if i >= j then Symbol t_symbol else Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to '>='. Both arguments must be numbers.\n");;
      
  
(*************** List processing *********************)
  
(* function 'cons' *)
let builtin_cons args =
  match args with
    [|car ; cdr|] -> Cons (car, cdr)
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to 'cons'. Both arguments must be either lists or atomic.\n");;
  
(* function 'car' *)
let builtin_car args =
  match args with
    [|Cons (car, _)|] -> car
  | _ -> raise (Builtin_Invalid_Args "Invalid argument to 'car'. This argument must be a list.\n");;
  
(* function 'cdr' *)
let builtin_cdr args =
  match args with
    [|Cons (_, cdr)|] -> cdr
  | _ -> raise (Builtin_Invalid_Args "Invalid argument to 'cdr'. This argument must be a list.\n");;

(* function 'null?' *)
let builtin_nul args =
  match args with
    [| Nil |] -> Symbol t_symbol
  |[| Cons (car, cdr) |] -> Nil
  | _ -> raise (Builtin_Invalid_Args "Invalid argument to 'null?'. This argument must be a list.\n");; 


(*************** Strings *********************)

(* function 'string-length' *) 
let builtin_stringlength args =
  match args with
  	[|String(s)|] -> Int((String.length s))
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to 'string-length'. Usage: (string-length <string>)\n");;

(* function 'substring' *) 
let builtin_substring args =
  match args with
  	[|String(s); Int(s1)|] ->  let ssub = try String.sub s s1 ((String.length s) - s1) with _ -> "" in String(ssub)
  |	[|String(s); Int(s1); Int(s2)|] -> let ssub = try String.sub s s1 (s2 - s1) with _ -> "" in String(ssub)
  | _ -> raise (Builtin_Invalid_Args "Invalid arguments to 'substring'. Usage: (substring <string> <start position> <(optional) end position>)\n");;
  
(* function 'string-append'*)  
let builtin_stringappend args = 
  Array.fold_left (function x -> function y -> 
  						match (x, y) with 
  						(String(i), String(j)) -> String(i ^ j)
  					| _ -> raise (Builtin_Invalid_Args "Invalid arguments to 'string-append'. All arguments must be strings.\n")) (String("")) args;;

  
(*************** Other *********************)

(* function 'begin' *)  
let builtin_begin args = 
	let tonull x =  () in
	for i = 0 to (pred (Array.length args)) do
		tonull (compile args.(i) [])
	done;
	Symbol t_symbol;;



(* function 'display' *)
let builtin_display args =
	for i = 0 to (pred (Array.length args)) do
		prin1 args.(i)
	done;
	print_newline();
	flush stdout;	
	Symbol t_symbol;;

(**************************************************************************
	LOAD BUILTINS
**************************************************************************)	    
  
let main () =
  t_symbol.value <- Value (Symbol t_symbol) ;
  nil_symbol.value <- Value (Nil) ;
  
  (* arithmetic and logic *)
  (intern "+").value <- Value (Builtin builtin_plus) ;
  (intern "-").value <- Value (Builtin builtin_minus) ;
  (intern "*").value <- Value (Builtin builtin_times) ;
  (intern "/").value <- Value (Builtin builtin_divide) ;
  (intern "quotient").value <- Value (Builtin builtin_quotient) ;
  (intern "modulo").value <- Value (Builtin builtin_modulo) ;
  (intern "=").value <- Value (Builtin builtin_numbers_equal) ;
  (intern "not").value <- Value (Builtin builtin_numbers_not) ;
  (intern "<").value <- Value (Builtin builtin_less) ;
  (intern ">").value <- Value (Builtin builtin_more) ;
  (intern "<=").value <- Value (Builtin builtin_lesseq) ;
  (intern ">=").value <- Value (Builtin builtin_moreeq) ;
  (* lists*)
  (intern "cons").value <- Value (Builtin builtin_cons) ;
  (intern "car").value <- Value (Builtin builtin_car) ;
  (intern "cdr").value <- Value (Builtin builtin_cdr) ;
  (intern "null?").value <- Value (Builtin builtin_nul) ;
  (* strings *)
  (intern "string-length").value <- Value (Builtin builtin_stringlength) ;
  (intern "substring").value <- Value (Builtin builtin_substring) ;
  (intern "string-append").value <- Value (Builtin builtin_stringappend) ;
  (* other *)
  (intern "begin").value <- Value (Builtin builtin_begin) ;
  (intern "display").value <- Value (Builtin builtin_display) ;
  let lexbuf = Lexing.from_channel stdin in
  try 
  while true do
	    let expr = macroexpand (Parser.expr Lexer.token lexbuf) in
    	prin1 (bceval (compile expr []) []); print_newline(); flush stdout
  done
  with
    Lexer.Eof -> Printf.printf "Done"
  | Builtin_Invalid_Args s -> Printf.printf "Line %d: %s\n" (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1) s
  | Apply_Error -> Printf.printf "Line %d: Error applying expression\n"  (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1)
  | Print_Error -> Printf.printf "Line %d: Error printing expression\n"  (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1)
  | Not_Symbol -> Printf.printf "Line %d: Invalid or unknown symbol\n"  (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1)
  | Not_Value -> Printf.printf "Line %d: Invalid or unknown value\n"  (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1)
  | Compile_Error -> Printf.printf "Line %d: Error compiling expression\n"  (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1)
  | Expand_Error -> Printf.printf "Line %d: Error expanding expression\n"  (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1)
  | _ -> Printf.printf "Line %d: Error\n" (lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum + 1);;

main();;
exit 0;;
